package cs206proj.part3;

/**
 * A hash table implementation of the Dictionary interface that uses 
 * chaining to resolve collisions    
 * The key field of the (element, key) pair must override the hashCode()
 * method of the Object class and return its 32 bit integer hashcode, and
 * must override the equals method of the Object class
 * @author Tia Newhall, Eric Eaton
 * 
 * @date  11/14/11 - Eric Eaton:  converted to generics
 *
 * Copyright (c) 2000 Tia Newhall, Lisa Meeden
 */
import java.util.*;

public class ChainingHashtable<K,V> implements Dictionary<K,V> {

	private LinkedList<KeyElementPair<K,V>>[] hashTable;  // list of linked lists of KeyElementPair 
	private int 	 size;     
	private static int 	 N = 51;  // default num_buckets     

	private class KeyElementPair<K,V> {
		private K key;
		private V element;
		KeyElementPair(K k, V e) { element = e; key = k; }
		public K getKey() { return key; }	
		public V getElement() { return element; }	
		public boolean equals(Object k) { return key.equals(k); }
		public boolean equals(KeyElementPair<K,V> obj) { return key.equals(obj.getKey()); }
		public String toString() { 
			return("(" + element + ", " + key + ")\t hash code = " 
					+ key.hashCode());
		}

	}

	/**
	 * Create a new hastable with given number of buckets  
	 * @param buckets the number of hashtable buckets
	 */
	@SuppressWarnings("unchecked")
	public ChainingHashtable(int buckets) {
		hashTable = new LinkedList[buckets];	
		for(int i=0; i < buckets; i++)
			hashTable[i] = new LinkedList<KeyElementPair<K,V>>();
		size = 0;
	}
	/**
	 * Create a new hashtable with 51 buckets  
	 */
	public ChainingHashtable() { this(N); }

	// map hash code of a key to a bucket number
	private int keyToBucketNumber(K key) { 
		return( (Math.abs(key.hashCode())) % hashTable.length ); 
	} 

	/**
	 * Returns an object with a matching key
	 * @param key the key value to find
	 * @return an object with a matching key
	 */
	public V get(K key) {
		if(size == 0) return null;
		int bucket = keyToBucketNumber(key);
		int len = hashTable[bucket].size();
		for(int i=0; i < len; i++) {
			KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[bucket].get(i);  
			if(next.equals(key)) {
				return (next.getElement());
			}
		}
		return null;
	} 

	/**
	 * Add a new (element,key) pair to the dictionary
	 * @param key the key value used to search for the element
	 * @param element the data value to store
	 */
	public void put(K key, V element) {
		int bucket = keyToBucketNumber(key);
		hashTable[bucket].add( (new KeyElementPair<K,V>(key, element)) );	
		size++;
	}

	/**
	 * Returns an object with a matching key and remove it from the Dictionary
	 * @param key the key value to find
	 * @return an object with a matching key that was removed, or null if no 
	 * match
	 */
	public V remove(K key) {
		if(size == 0) return null;
		int bucket = keyToBucketNumber(key);
		int len = hashTable[bucket].size();
		for(int i=0; i < len; i++) {
			KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[bucket].get(i);  
			if(next.equals(key)) {
				next = (KeyElementPair<K,V>)hashTable[bucket].remove(i);  
				size--;
				return next.getElement();
			}
		}
		return null;
	}

	/**
	 * Returns an Iterator over all dictionary objects with a matching key 
	 * @param key the key value to find
	 * @return an object with a matching key
	 */
	public Iterator<V> getAll(K key) {
		LinkedList<V> list = new LinkedList<V>();
		int bucket = keyToBucketNumber(key);
		int len = hashTable[bucket].size();
		for(int i=0; i < len; i++) {
			KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[bucket].get(i);  
			if(next.equals(key)) {
				list.add(next.getElement()); 
			}
		}
		return (new DictionaryIterator<V>(list));	    
	} 

	/**
	 * Returns an Iterator over all dictionary objects with a matching key 
	 * and removes them from the dictionary
	 * @param key the key value to find
	 * @return an Iterator of all objects that were removed or null if no match
	 */
	public Iterator<V> removeAll(K key) {
		LinkedList<V> list = new LinkedList<V>();
		int bucket = keyToBucketNumber(key);
		int len = hashTable[bucket].size();
		int i=0;
		for(int j=0; j < len; j++) {
			KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[bucket].get(i);
			if(next.equals(key)) {
				next = (KeyElementPair<K,V>)hashTable[bucket].remove(i);
				size--;
				list.add(next.getElement());
			} else {
				i++;
			}
		}
		return (new DictionaryIterator<V>(list));
	}
	/**
	 * Returns the number of elements in the dictionary
	 * @return the number of elements in the dictionary
	 */
	public int size() { return size; }

	/**
	 * Returns true if this dictionary is empty
	 * @return true if this dictionary is empty
	 */
	public boolean isEmpty() { return (size == 0); } 

	/**
	 * Returns true if dictionary contains an object with the specified key
	 * @param key the key to search for
	 * @return true if dictionary contains an object with the specified key
	 */
	public boolean containsKey(K key) { return(this.get(key) != null); } 

	/**
	 * Returns an iterator over all keys
	 * @return an iterator over all keys
	 */
	public Iterator<K> keys() {
		LinkedList<K> list = new LinkedList<K>();
		for(int bucket = 0; bucket < hashTable.length; bucket++) { 
			int len = hashTable[bucket].size();
			for(int i=0; i < len; i++) {
				KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[bucket].get(i);  
				list.add(next.getKey()); 
			}
		}
		return (new DictionaryIterator<K>(list));	    
	} 

	/**
	 * Returns an iterator over all elements
	 * @return an iterator over all elements
	 */
	public Iterator<V> elements() {
		LinkedList<V> list = new LinkedList<V>();
		for(int bucket = 0; bucket < hashTable.length; bucket++) { 
			int len = hashTable[bucket].size();
			for(int i=0; i < len; i++) {
				KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[bucket].get(i);  
				list.add(next.getElement()); 
			}
		}
		return (new DictionaryIterator<V>(list));	    
	} 

	/**
	 * Returns true if this dictionary contains an matching object 
	 * @param element the object to find
	 * @return true if this dictionary contains an matching object 
	 */
	public boolean containsElement(Object element) { 
		DictionaryIterator<V> elms = (DictionaryIterator<V>)elements();
		while(elms.hasNext()) {
			if(elms.next().equals(element))
				return true;
		}
		return false;   

	} 

	/**
	 * Convert a hash table to a string
	 */
	public String toString() {
		int i, j;
		String str = "Hash Table Dictionary\n---------------------\n";	
		str += "Size: " + size + "\n";
		for(i=0; i < hashTable.length; i++) {
			str = str + i + ":\n";
			for(j=0; j < hashTable[i].size(); j++) {
				KeyElementPair<K,V> next = (KeyElementPair<K,V>)hashTable[i].get(j);  
				str = str + "  " + next + "\n";
			}
		}
		return str;
	}
}
