package cs206proj.part3;

/**
 * An open addressing hash table implementation of the Dictionary interface 
 * that uses linear probing to resolve collisions    
 * The key field of the (element, key) pair must override the hashCode()
 * method of the Object class and return its 32 bit integer hashcode, and
 * must override the equals method of the Object class
 *
 * @date  11/14/11 - Eric Eaton:  converted to generics
 *
 *
 * Copyright (c) 2000 Tia Newhall, Lisa Meeden
 */
import java.util.*;

public class OpenAddressingHashtable<K,V> implements Dictionary<K,V> {
	private KeyElementPair<K,V>[] hashTable;  
	private int 	 size;     
	private static int 	 N = 51;  // default num_buckets     
	private KeyElementPair<K,V> removed; 

	private class KeyElementPair<K,V> {
		private K key;
		private V element;
		KeyElementPair(K k, V e) { element = e; key = k; }
		public K getKey() { return key; }	
		public V getElement() { return element; }	
		public boolean equals(Object o) { return key.equals(o); }
		public boolean equals(KeyElementPair<K,V> obj) { return key.equals(obj.getKey()); }
		public String toString() {
			return("(" + element + ", " + key + ")\thash code = "
					+ key.hashCode());
		}
	}

	/**
	 * Create a new hastable with given number of buckets  
	 * @param buckets the number of hashtable buckets
	 */
	@SuppressWarnings("unchecked")
	public OpenAddressingHashtable(int buckets) {
		hashTable = new KeyElementPair[buckets];	
		removed = new KeyElementPair<K,V>(null, null);
		size = 0;
	}

	/**
	 * Create a new hashtable with 51 buckets  
	 */
	public OpenAddressingHashtable() { this(N); }

	// map hash code of a key to a bucket number
	private int KeyToBucketNumber(Object key) { 
		return( (Math.abs(key.hashCode())) % hashTable.length ); 
	} 

	// search for a matching element based on the key value
	// using linear probing if there are collisions
	// return -1 if no match, else return the match's bucket index 
	private int findMatchingBucketNumber(Object key) {
		int bucket = KeyToBucketNumber(key);		
		for(int i=0; i < hashTable.length; i++) {
			KeyElementPair<K,V> next = hashTable[((bucket+i)%hashTable.length)];  
			if(next == null) {
				return -1;
			} else if(next.equals(key)) {
				return ((bucket+i)%hashTable.length);
			} 
		}
		return -1;
	}

	// search for a hole in the hashtable
	// either h(key) or first hole found by linear probing from h(key) 
	//
	private int findFirstHole(Object key) {
		int bucket = KeyToBucketNumber(key);		
		for(int i=0; i < hashTable.length; i++) {
			int index = (bucket+i)%hashTable.length;
			KeyElementPair<K,V> next = hashTable[index];  
			if(next == null) {
				return (index);
			} else if(next.getKey() == null) {
				return (index);
			} 
		}
		return -1;
	}

	/**
	 * Returns an object with a matching key
	 * @param key the key value to find
	 * @return an object with a matching key
	 */
	public V get(K key) {
		if(size == 0) return null;
		int found = findMatchingBucketNumber(key);
		if(found == -1) { return null; }
		return (hashTable[found].getElement());
	} 

	/**
	 * Add a new (element,key) pair to the dictionary
	 * @param key the key value used to search for the element
	 * @param element the data value to store
	 */
	public void put(K key, V element) {
		int bucket = findFirstHole(key);
		if(bucket != -1) {
			hashTable[bucket] = new KeyElementPair<K,V>(key, element);	
			size++;
		}
	}

	/**
	 * Returns an object with a matching key and remove it from the Dictionary
	 * @param key the key value to find
	 * @return an object with a matching key that was removed, or null if no 
	 * match
	 */
	public V remove(K key) {
		if(size == 0) return null;
		int found = findMatchingBucketNumber(key);
		if(found == -1) { return null; }
		V gone = hashTable[found].getElement();
		hashTable[found] = removed;
		size--;
		return gone;
	}

	/**
	 * Returns an Iterator over all dictionary objects with a matching key 
	 * @param key the key value to find
	 * @return an object with a matching key
	 */
	public Iterator<V> getAll(K key) {
		LinkedList<V> list = new LinkedList<V>();
		int index = findMatchingBucketNumber(key);
		if (index != -1)
			for (int i=0; i<hashTable.length; i++) {
				if (hashTable[index] == null) break;
				if (hashTable[index].equals(key)) { 
					list.add(hashTable[index].getElement());
				}
				index = (index + 1)%hashTable.length;
			}
		return (new DictionaryIterator<V>(list));
	} 

	/**
	 * Returns an Iterator over all dictionary objects with a matching key 
	 * and removes them from the dictionary
	 * @param key the key value to find
	 * @return an Iterator of all objects that were removed or null if no match
	 */
	public Iterator<V> removeAll(K key) {
		LinkedList<V> list = new LinkedList<V>();
		V next = this.remove(key);			
		while(next != null) {
			list.add(next);
			next = this.remove(key);			
		}
		return (new DictionaryIterator<V>(list));	    
	}


	/**
	 * Returns the number of elements in the dictionary
	 * @return the number of elements in the dictionary
	 */
	public int size() { return size; }

	/**
	 * Returns true if this dictionary is empty
	 * @return true if this dictionary is empty
	 */
	public boolean isEmpty() { return (size == 0); } 

	/**
	 * Returns true if dictionary contains an object with the specified key
	 * @param key the key to search for
	 * @return true if dictionary contains an object with the specified key
	 */
	public boolean containsKey(K key) { return(this.get(key) != null); } 

	/**
	 * Returns an iterator over all keys
	 * @return an iterator over all keys
	 */
	public Iterator<K> keys() {
		LinkedList<K> list = new LinkedList<K>();
		for(int i=0; i < hashTable.length; i++) {
			if((hashTable[i] != null) && (hashTable[i].getKey() != null)) {
				list.add(hashTable[i].getKey());
			}
		}
		return (new DictionaryIterator<K>(list));	    
	} 

	/**
	 * Returns an iterator over all elements
	 * @return an iterator over all elements
	 */
	public Iterator<V> elements() {
		LinkedList<V> list = new LinkedList<V>();
		for(int i=0; i < hashTable.length; i++) {
			if((hashTable[i] != null) && (hashTable[i].getKey() != null)) {
				list.add(hashTable[i].getElement());
			}
		}
		return (new DictionaryIterator<V>(list));	    
	} 

	/**
	 * Returns true if this dictionary contains an matching object 
	 * @param element the object to find
	 * @return true if this dictionary contains an matching object 
	 */
	public boolean containsElement(V element) { 
		DictionaryIterator<V> elms = (DictionaryIterator<V>)elements();
		while(elms.hasNext()) {
			if(elms.next().equals(element))
				return true;
		}
		return false;   
	} 

	/**
	 * convert a hashtable to a String 
	 * @return the hashtable as a String
	 */
	public String toString() {
		String str = "Hash Table Dictionary\n---------------------\n";
		str += "Size: " + size + "\n";
		for(int i=0; i < hashTable.length; i++) {
			str += i + ":\t";
			if (hashTable[i] == removed)
				str += "REMOVED\n";
			else
				str += hashTable[i] + "\n";
		}
		return str;
	}

}
